# SPDX-FileCopyrightText: 2022 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from flict.flictlib.lic_comp import LicenseCompatibilty
from flict.flictlib.project.reader import Project
from flict.flictlib.utils import meta_information
from flict.flictlib.utils import timestamp
from flict.flictlib.return_codes import FlictError, ReturnCodes
from flict.flictlib.license import compatible_license_short
from flict.flictlib.project.reader import FlictProjectReader


class Arbiter:
    """Arbiter is a class to verify compatibility"""

    def __init__(self, license_db=None, licenses_preferences=None, denied_licenses=None, update_dual=True):
        """Initializes Arbiter objects
             Parameters:
                 license_db: license database to use instead of builtin
                 licenses_preferences: license preferences to use instead of builtin
                 denied_licenses: licenses that cannot be used
        """
        self.update_dual = update_dual
        self.license_compatibility = LicenseCompatibilty(
            license_db=license_db, licenses_preferences=licenses_preferences, denied_licenses=denied_licenses, update_dual=update_dual)

    def supported_licenses(self):
        """Returns the supported licenses"""
        return self.license_compatibility.supported_licenses()

    def _verify_package(self, package, licenses):
        """Verifies a package's license to a list of outbounds and returns the
        compatibility between the liceses.
             Parameters:
                 package: the package (with its license) to check for compatibility
                 licenses: the licenses to check the package's license against
        """
        logging.debug(f"* verify package {package['name']} (\"{package['license']}\"")

        logging.debug(f"   * All licenses: {licenses}")
        inbound_license = package['license']
        logging.debug(f"   * Inbound license:  {inbound_license}")

        checks = []
        problems = []
        for outbound_license in licenses:
            check = self.inbounds_outbound_check(outbound_license, [inbound_license])
            problems += check.get('problems', [])
            checks.append(check)

        return checks, problems

    def verify_outbound_inbound(self, outbound, inbound):
        # Use that flict can calculate license compatibility for a
        # project by creating an dummy project - with outer project
        # (outbound license) and one simple dependency (inbound
        # license).
        dummy_project_data = {
            'project': {
                'name': '',
                'version': '',
                'license': ' '.join(outbound),
                'dependencies': [
                    {
                        'name': '',
                        'version': '',
                        'license': ' '.join(inbound),
                        'dependencies': [],
                    },
                ],
            },
        }
        reader = FlictProjectReader([])
        dummy_project = reader.read_project_data(dummy_project_data)
        compat_outbound = compatible_license_short(' '.join(outbound), self.update_dual)
        verification = self.verify(dummy_project, [compat_outbound])

        package = verification['packages'][0]
        dependency = package['dependencies'][0]
        problems = package.get('problems', []) + dependency['problems']
        allowed_outbounds = package.get('allowed_outbound_licenses', [])
        return {
            'original_outbound': package['original_license'],
            'outbound': package['license'],
            'inbound': dependency['license'],
            'original_inbound': dependency['original_license'],
            'result': {
                'outbound_licenses': package.get('outbound_licenses', []),
                'allowed_outbound_licenses': allowed_outbounds,
                'outbound_license': package.get('outbound_license') or '',
                'problems': problems,
            },
        }

    def _package_info_compatibility(self, package_info, outbound):
        for compat in package_info['compatibility']:
            logging.debug(f"            *** outbound: \"{compat['outbound']}\"  <--->  \"{compat}\"")

            if compat['outbound']['name'] == outbound:
                logging.debug("         *** found outbound in compats")

                allowed = compat['allowed']
                compatibility = compat['compatibility']
                compatible = compatibility == "Yes"
                name = compat['name']
                logging.debug(f"         *** type:          {type}")
                logging.debug(f"         *** allowed:       {allowed}")
                logging.debug(f"         *** compatibility: {compatibility}")
                logging.debug(f"         *** compatible:    {compatible}")
                logging.debug(f"         *** name:          {name}")

                return (compatible, compat['name'])

        return (False, None)

    def _combined_work_compatible_dependencies(self, outbound_license, dep_infos):
        compatible = True

        for dep_info in dep_infos:
            # Get the corresponding compatiblity for outbound_license for dep_info
            compatible_license = self._package_info_compatibility(dep_info, outbound_license)
            dep_compatible = compatible_license[0]
            compatible = compatible and dep_compatible
            logging.debug("        {dep_info['name']}: {dep_compatible} ===> {compatible}")
        return compatible

    def _combined_work_compatible(self, outbound_license, package_info, dep_infos):
        compatible_license = self._package_info_compatibility(package_info, outbound_license)
        if not compatible_license:
            raise FlictError(ReturnCodes.RET_INTERNAL_ERROR,
                             f"Found license incompatibility for \"{package_info}\", \"{outbound_license}\"")
        else:
            compatible = compatible_license[0]

        # AND together compatible status from dependencies
        compatible = compatible and self._combined_work_compatible_dependencies(outbound_license, dep_infos)

        logging.debug(f"    --> combined work compatible: {compatible}\n")
        return compatible

    def _top_package_license(self, all_licenses, package_info, dep_infos):
        outbound_licenses = set()
        logging.debug("TOP LEVEL OUTBOUND")
        #
        # Start with checking top package's license as outbound
        #
        compatible = self._combined_work_compatible(package_info['license'], package_info, dep_infos)
        if compatible:
            outbound_licenses.add(package_info['license'])

        #
        # Let's loop through all the other licenses in the combined work
        #
        for lic in all_licenses:
            compatible = self._combined_work_compatible(lic, package_info, dep_infos)
            if compatible:
                outbound_licenses.add(lic)

        return list(outbound_licenses)

    def _package_info(self, package, licenses):
        compats, problems = self._verify_package(package, licenses)
        return {
            'name': package['name'],
            'license': package.get('license'),
            'original_license': package.get('original_license'),
            'license_details': package.get('license_details', ""),
            'licenses_to_check': list(licenses),
            'version': package['version'],
            'description': package.get('description', ""),
            'compatibility': compats,
            'problems': list(set(problems)),
        }

    def verify(self, project, supplied_licenses=None):
        """Verifies a project's license to a list of outbounds and returns the
        compatibility between the liceses.
             Parameters:
                 project: the project (with its packages and their licenses) to check for compatibility
                 supplied_licenses: the licenses to check the package's license against
        """
        start_time = timestamp()

        project_name = project['project_name']

        package_infos = []
        all_licenses = set()

        for package in project['packages']:
            license_expression = Project.combined_work_license(package)

            if supplied_licenses is None:
                licenses = self.license_compatibility.licenses(license_expression)
            else:
                licenses = self.license_compatibility.licenses(' '.join(supplied_licenses))

            all_licenses.update(licenses)

            package_info = self._package_info(package, licenses)

            dep_infos = [self._package_info(dep, licenses) for dep in package.get('dependencies', [])]
            dep_problems = [problem for dep_info in dep_infos for problem in dep_info['problems']]

            # Get a list of the outbound licenses for all packages
            outbound_licenses = self._top_package_license(licenses, package_info, dep_infos)

            allowed_outbound_licenses = [ol for ol in outbound_licenses if self.license_compatibility.license.license_allowed(ol)]

            # Identify single outbound (chosen) license (from the aliased outbound licenses)
            chosen_license = self.license_compatibility.choose_license(allowed_outbound_licenses)

            package_info.update({
                'dependencies': dep_infos,
                'dependency_problems': list(set(dep_problems)),
                'outbound_licenses': outbound_licenses,
                'allowed_outbound_licenses': allowed_outbound_licenses,
                'outbound_license': chosen_license,
            })
            package_infos.append(package_info)

        return {
            "project_name": project_name,
            "packages": package_infos,
            "meta": meta_information(start_time),
            "all_licenses": list(all_licenses),
        }

    def _compat_ok_to_use(self, compat):
        """
        Checks if a compatibility is OK to use by checking if
        * the license is compatible
        * the license is not denied
        """
        return compat['compatibility'] == "Yes" and compat['allowed']

    def inbounds_outbound_check(self, outbound, expr):
        """Check an outbound license against inbound licenses
             Parameters:
                 outbound: the outbound license (e.g. "GPL-2.0-only")
                 expr: license expr with inbound license (e.g. "MPL-2.0 OR MIT")
        """
        return self.license_compatibility.inbounds_outbound_compatibility(outbound, expr)

    def inbound_outbound_check(self, outbound, inbound):
        """Check an outbound license against an inbound license
             Parameters:
                 outbound: the outbound license (e.g. "GPL-2.0-only")
                 inbound: inbound license (e.g. "MPL-2.0")
        """
        return self.license_compatibility.inbound_outbound_compatibility(outbound, inbound)

    def check_compatibilities(self, licenses, check_all=False):
        """Check compatbilitiy between supplied licenses"""
        return self.license_compatibility.check_compatibilities(licenses, check_all)

    def extend_license_db(self, file_name, oformat="JSON", default_no=False):
        return self.license_compatibility.extend_license_db(file_name, oformat, default_no)

    def simplify_license(self, expr):
        return self.license_compatibility.simplify_license(expr)

    def parse_license(self, expr):
        return self.license_compatibility.parse_license(expr)

    def licenses(self, expr):
        return self.license_compatibility.licenses(expr)

    def license_compatibility_as(self, expr):
        return compatible_license_short(expr)



# SPDX-FileCopyrightText: 2022 Henrik Sandklef
# SPDX-FileCopyrightText: 2022 Jens Erdmann
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import logging

from flict.flictlib.return_codes import FlictError, ReturnCodes
from flict.flictlib.license import compatible_license

DEPENDENCY_TAGS = ['DYNAMIC_LINK', 'STATIC_LINK', 'DEPENDS_ON', 'CONTAINS', 'COPY_OF']

class Project:

    def __init__(self, project_file, project_dirs):
        project_reader = ProjectReaderFactory.get_projectreader(project_file, project_dirs)
        self.project = project_reader.read_project()

    @staticmethod
    def dependencies_license(package):
        licenses = [f' ( {Project.package_license(dep)} ) ' for dep in package.get('dependencies', [])]
        return ' AND '.join(licenses)

    @staticmethod
    def package_license(package):
        if 'license' not in package:
            raise FlictError(ReturnCodes.RET_INVALID_PROJECT, f'Package does not contain "license": {package}')
        return package['license']

    @staticmethod
    def combined_work_license(package):
        if Project.dependencies_license(package).strip() != '':
            return f' ( {Project.package_license(package)} ) AND ( {Project.dependencies_license(package)} ) '
        else:
            return Project.package_license(package)


class ProjectReader:
    """Interface for classes that reads project from some kind of format"""

    def read_project(self):
        return

    def _add_files_read(self, project_file):
        self.files_read.append(project_file)

    def _already_read(self, project_file):
        return project_file in self.files_read

    def prepare_project(self, project):
        for package in project.get('packages', []):
            self.__update_dep_license(package)
        return project

    def __update_dep_license(self, package):
        package['original_license'] = package.get('license')
        compat = compatible_license(package.get('license'), update_dual=self.update_dual)
        fixed_license = compat['compat_license']
        package['license'] = fixed_license

        for dep in package['dependencies']:
            self.__update_dep_license(dep)

    def _flatten_packages(self, packages):
        package_list = []

        for _package in packages:
            tmp_dict = {}
            for dep in _package.get('dependencies', []):
                tmp_dict.update(self._flatten_package_tree(dep))
            dep_list = tmp_dict.values()

            for dep in dep_list:
                dep['dependencies'] = []

            package = {
                'name': _package['name'],
                'version': _package['version'],
                'license': _package['license'],
                'description': _package['description'],
                'dependencies': list(dep_list),
            }

            package_list.append(package)

        return package_list


class ProjectReaderFactory:

    @staticmethod
    def get_projectreader(project_file=None, project_dirs=None, project_format=None, update_dual=True):
        if not project_dirs:
            project_dirs = ['.']
        logging.debug(f'get_projectreader({project_file}, {project_dirs}, {project_format}, {update_dual})')
        if project_format is None:
            if 'spdx' in project_file:
                return SPDXJsonProjectReader(project_dirs, update_dual)
            if 'flict' in project_file:
                return FlictProjectReader(project_dirs, update_dual)
        elif project_format == 'flict':
            return FlictProjectReader(project_dirs, update_dual)
        elif project_format == 'spdx':
            return SPDXJsonProjectReader(project_dirs, update_dual)


class FlictProjectReader(ProjectReader):
    """Class for reading flict project files"""

    def __init__(self, project_dirs, update_dual=True):
        self.update_dual = update_dual

    def read_project_data(self, project_data):
        project = project_data['project']
        project_data = {
            'project_name': project['name'],
            'packages': [
                {
                    'name': project['name'],
                    'version': project['version'],
                    'license': project['license'],
                    'description': '',
                    'dependencies': project['dependencies'],
                },
            ],
        }
        fixed_project_data = self.prepare_project(project_data)
        return fixed_project_data

    def read_project(self, project_file):
        try:
            with open(project_file, 'r') as f:
                project_data = json.load(f)
                return self.read_project_data(project_data)
        except json.JSONDecodeError:
            raise FlictError(ReturnCodes.RET_INVALID_PROJECT, f'File "{project_file}" does not contain valid JSON data')
        except (FileNotFoundError, IsADirectoryError):
            raise FlictError(ReturnCodes.RET_FILE_NOT_FOUND, f'File "{project_file}" could not be found or is a directory')
        except Exception as e:
            raise FlictError(ReturnCodes.RET_INVALID_PROJECT,
                             f'File "{project_file}" could not be parsed, probably not a flict project file: {e}')


class SPDXJsonProjectReader(ProjectReader):
    """Class for reading SBoM files in SPDX 2.2
    If more SPDX versions need to be supported, we may need to subclass this in version specific classes....
    """

    def __init__(self, project_dirs, update_dual=True):
        self.spdx_dirs = project_dirs
        self.files_read = []
        self.update_dual = update_dual

    def read_project(self, project_file):
        packages_project_name = self._read_spdx(project_file)
        packages = packages_project_name['packages']
        project_name = packages_project_name['project_name']
        flat_packages = self._flatten_packages(packages)

        fixed_project_data = self.prepare_project({
            'project_name': project_name,
            'packages': flat_packages,
        })
        return fixed_project_data

    def _relationship_is_dependency(self, relationshiptype):
        return relationshiptype in DEPENDENCY_TAGS

    def _read_spdx_2_2(self, only_packages=None):
        project_name = self.project['name']
        packages = {}
        for pkg in self.project['packages']:
            elem_id = pkg['SPDXID']
            packages[elem_id] = {
                'id': elem_id,
                'name': pkg['name'],
                'version': pkg.get('versionInfo', ''),
                'license': pkg['licenseConcluded'],
                'description': pkg['description'],
                'dependencies': [],
            }

        for dep in self.project.get('relationships', []):
            top_package = dep['relatedSpdxElement']
            relationshiptype = dep['relationshipType']
            if not self._relationship_is_dependency(relationshiptype):
                logging.debug(f"relationship {dep.get('spdxElementId')} ignored since {relationshiptype} is not defined to be a dependency tag.")
                continue

            if 'DocumentRef' in dep['spdxElementId'] and ':' in dep['spdxElementId']:
                dep_package_doc = dep['spdxElementId'].split(':')[0]
                dep_package_name = dep['spdxElementId'].split(':')[1]
                dep_spdx = dep_package_doc.replace('DocumentRef-', '') + '.spdx.json'
                if self._already_read(dep_spdx):
                    continue
                self._add_files_read(dep_spdx)
                dep_spdx_path = self.spdx_dirs[0] + '/' + dep_spdx
                packages_proj_name = self._read_spdx(dep_spdx_path, self.spdx_dirs)
                _packages = packages_proj_name.get('packages')
                for _pkg in _packages:
                    _pkg_name = _pkg['id']
                    if dep_package_name == _pkg_name:
                        packages[top_package]['dependencies'].append({_pkg_name: _pkg})
            else:
                dep_id = dep['spdxElementId']
                for package in self.project['packages']:
                    if dep_id == package['SPDXID']:
                        if top_package not in packages:
                            logging.warning(f'package {top_package} marked as a relationship, but without corresponding package definition.')
                        else:
                            _pkg = {
                                'id': dep_id,
                                'name': pkg['name'],
                                'version': pkg.get('versionInfo', ''),
                                'license': pkg['licenseConcluded'],
                                'description': pkg['description'],
                                'dependencies': [],
                            }
                            packages[top_package]['dependencies'].append({dep_id: _pkg})

        packages_list = list(packages.values())

        ret = {
            'packages': packages_list,
            'project_name': project_name,
        }
        return ret

    def _read_spdx(self, spdx_file, only_packages=None):

        try:
            with open(spdx_file, 'r') as f:
                self.project = json.load(f)

        except json.JSONDecodeError:
            raise FlictError(
                ReturnCodes.RET_INVALID_PROJECT,
                f'File "{spdx_file}" does not contain valid JSON data',
            )

        spdx_version = self.project['spdxVersion'].replace('SPDX-', '')

        if spdx_version.startswith('2.2'):
            return self._read_spdx_2_2(only_packages=None)
        raise FlictError(ReturnCodes.RET_INTERNAL_ERROR,
                         f'SPDX version ({spdx_version}) not supported.')

    def _flatten_package_tree(self, packages):
        package_dict = {}

        for _package in packages.values():
            package_dict.update(
                {f"{_package['name']}--{_package['version']}": _package},
            )
            for dep in _package.get('dependencies', []):
                package_dict.update(self._flatten_package_tree(dep))

        return package_dict



###################################################################
#
# flict - FOSS License Compatibility Tool
#
# SPDX-FileCopyrightText: 2020 Henrik Sandklef, 2021 Konrad Weihmann
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
###################################################################

import json
import os

from osadl_matrix import OSADL_MATRIX_JSON


def read_user_config():
    for path in [
        os.environ.get('FLICT_USERCONFIG'),
        os.path.join(os.environ.get('HOME', '/does/not/exist'), '.flict.cfg'),
    ]:
        try:
            with open(path) as i:
                return json.load(i)
        except Exception:  # noqa: S110 - it's okay ignore all errors here
            pass
    return {}


_userconfig = read_user_config()

flict_version = "1.2.8"

SCRIPT_DIR = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../"))

VAR_DIR = os.path.join(SCRIPT_DIR, "var")

DEFAULT_MATRIX_FILE = _userconfig.get('matrix-file', OSADL_MATRIX_JSON)
DEFAULT_OUTPUT_FORMAT = _userconfig.get('output-format', "JSON")



# SPDX-FileCopyrightText: 2022 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later


import logging
import license_expression

from flict.flictlib.return_codes import FlictError, ReturnCodes

from enum import Enum


class LicenseExpression(Enum):
    LICENSE_EXPRESSION_OR = "OR"
    LICENSE_EXPRESSION_AND = "AND"


class PrettyLicenseSymbol(Enum):
    LICENSE_SYMBOL = "LicenseSymbol"
    LICENSE_WITH_SYMBOL = "LicenseWithExceptionSymbol"


class LicenseParserFactory:

    @staticmethod
    def get_parser():
        # Not much of a choice really :)
        return PrettyLicenseParser()


class LicenseParser:

    def __init__(self):
        self.utils = ParseUtils()

    def parse_license(self, expr):
        return

    def licenses(self, expr):
        return

    def is_operator(self, expr):
        return expr['type'] == "operator"

    def is_license(self, expr):
        return expr['type'] == "license"

    def operator(self, expr):
        return expr['name']

    def is_or(self, expr):
        return self.expr_operator(expr) == LicenseExpression.LICENSE_EXPRESSION_OR.value

    def is_and(self, expr):
        return self.operator(expr) == LicenseExpression.LICENSE_EXPRESSION_AND.value

    def license(self, expr):  # noqa: A003
        return expr['name']

    def operands(self, expr):
        return expr['operands']


class PrettyLicenseParser(LicenseParser):

    def __init__(self):
        super(PrettyLicenseParser, self).__init__()
        self.licensing = license_expression.Licensing()

    def parse_license(self, expr):
        logging.debug(f"parse_license(\"{expr}\")")

        if not isinstance(expr, list):
            raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION,
                             f"Internal error: Expression to parse must be a list: {expr}")

        if len(expr) == 0:
            raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION, f"Internal error: Expression list cannot be empty: {expr}")

        if len(expr[0]) == 0:
            raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION,
                             "Internal error: Expression list's first item cannot be empty")

        lic_expr = " ".join(expr).replace(")", " ) ").replace("(", " ( ")
        logging.debug(f"expression:         {lic_expr}")

        parsed = self.licensing.parse(lic_expr)
        logging.debug(f"parsed expression:  {parsed}")

        simplified = parsed.simplify()
        logging.debug(f"parsed expression:  {simplified}")

        pretty = self.licensing.parse(simplified).pretty()
        logging.debug(f"pretty expression:  {pretty}")

        trimmed = pretty.replace("\n", "")
        logging.debug(f"trimmed expression: {trimmed}")

        logging.debug(f"parse_license(\"{expr}\") => {trimmed}")
        return {
            'license': self._parse_license(trimmed),
            'original': lic_expr,
            'simplified': simplified,
            'pretty': pretty,
            'trimmed': trimmed,
        }

    def _parse_license(self, expr):
        logging.info(f"_parse_license(\"{expr}\")")

        is_operator = self.utils.is_operator(expr)

        if is_operator:

            op = self.utils.next_operator(expr)
            rest = expr[len(op):]

            # find last parenthesis and
            rest = rest[1:self.utils.index_last_parenthesis(rest)].strip()

            operands = []

        else:
            rest = expr

        while (rest != ""):
            if self.utils.is_license(rest):
                lic, rest = self.utils.get_license(rest)
                if lic == '':
                    continue
                operand = {
                    'type': 'license',
                    'name': lic,
                }

                if (is_operator):
                    operands.append(operand)
                else:
                    return operand
            elif self.utils.is_operator(rest):
                operation, rest = self.utils._get_op_expr(rest)
                logging.info(f"  operation: {operation}")
                parsed = self._parse_license(operation)
                operands.append(parsed)
            elif self.utils.next_token(rest) == ",":
                rest = self.utils.remove_comma(rest)
            elif self.utils.next_token(rest) == PrettyLicenseSymbol.LICENSE_WITH_SYMBOL.value:
                lic, rest = self.utils.get_license(rest)
            else:
                logging.error("*** PANIC IN DETROIT ***")
                logging.error(rest)
                raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION,
                                 f"Internal error: Remaining expression not valid: {rest}")
        return {
            'type': 'operator',
            'name': op,
            'operands': operands,
        }

    def licenses(self, expr):
        """
        Given a license expression, the licenses in the expression is returned as a list.

        Parameters:
            expr - license expression

        Examples:
          licenses("( FTL OR GPL-2.0-or-later ) AND (  ( Libpng)  AND  ( Zlib)  ) ")

          will return:
            'GPL-2.0-or-later', 'Zlib', 'FTL', 'Libpng'
        """
        try:
            logging.debug(f"licenses(\"{expr}\")")
            parsed = self.licensing.parse(expr.replace(" WITH ", "_WITH_"))
            keys = self.licensing.license_keys(parsed)
            key_set = set()
            for key in keys:
                key_set.add(key.replace("_WITH_", " WITH "))
            return list(key_set)
        except Exception:
            raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION, f"Could not parse and list license expression: {expr}")


class ParseUtils:

    def _find_expr_end(self, expr):
        logging.debug(f"_find_expr_end(\"{expr}\")")

        first_found = False
        paren_count = 0
        size = 0
        for tok in expr:
            if tok == "(":
                paren_count += 1
                first_found = True
            elif tok == ")":
                paren_count -= 1

            size += 1
            if paren_count == 0 and first_found:
                index = 0
                return size + index

        raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION, f"Internal error: Could not find end of expression: {expr}")

    def next_token(self, expr):
        for i in range(len(expr)):
            if i == " " or i == "\n":
                pass
            else:
                if expr.startswith(LicenseExpression.LICENSE_EXPRESSION_OR.value):
                    return LicenseExpression.LICENSE_EXPRESSION_OR.value
                elif expr.startswith(LicenseExpression.LICENSE_EXPRESSION_AND.value):
                    return LicenseExpression.LICENSE_EXPRESSION_AND.value
                elif expr.startswith(PrettyLicenseSymbol.LICENSE_SYMBOL.value):
                    return PrettyLicenseSymbol.LICENSE_SYMBOL.value
                elif expr.startswith(PrettyLicenseSymbol.LICENSE_WITH_SYMBOL.value):
                    return PrettyLicenseSymbol.LICENSE_WITH_SYMBOL.value

                return expr[i]

    def next_operator(self, expr):
        token = self.next_token(expr)
        if token == LicenseExpression.LICENSE_EXPRESSION_OR.value or token == LicenseExpression.LICENSE_EXPRESSION_AND.value:
            return token
        return None

    def is_operator(self, expr):
        return self.next_operator(expr) is not None

    def is_license(self, expr):
        return expr.startswith(PrettyLicenseSymbol.LICENSE_SYMBOL.value) or expr.startswith(PrettyLicenseSymbol.LICENSE_WITH_SYMBOL.value)

    def is_with(self, expr):
        return expr.startswith(PrettyLicenseSymbol.LICENSE_WITH_SYMBOL.value)

    def remove_comma(self, expr):
        stripped = expr.strip()
        return stripped[1:].strip()

    def _get_op_expr(self, expr):
        expr = expr.strip()
        index = self._find_expr_end(expr)
        operation = expr[:index]
        rest = expr[index:]
        return (operation, rest)

    def get_license(self, expr):
        index = expr.index(")")
        if PrettyLicenseSymbol.LICENSE_WITH_SYMBOL.value in expr:
            lic = expr[len(PrettyLicenseSymbol.LICENSE_WITH_SYMBOL.value) + 1:index].replace("'", "")
        elif PrettyLicenseSymbol.LICENSE_SYMBOL.value in expr:
            lic = expr[len(PrettyLicenseSymbol.LICENSE_SYMBOL.value) + 1:index].replace("'", "")

        rest = expr[index + 1:]
        return (lic, rest)

    def index_last_parenthesis(self, expr):
        return expr.rfind(")")



# SPDX-FileCopyrightText: 2022 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime
import os


def timestamp():
    return str(datetime.datetime.now())


def meta_information(start_time=""):
    uname = os.uname()
    return {
        'os': uname.sysname,
        'osRelease': uname.release,
        'osVersion': uname.version,
        'machine': uname.machine,
        'host': uname.nodename,
        'user': os.environ.get('USER'),
        "start_time": start_time,
        'stop_time': timestamp(),
    }



# SPDX-FileCopyrightText: 2021 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

from argparse import RawTextHelpFormatter
import argparse
import traceback

from flict.flictlib import logger
from flict.flictlib.flict_config import flict_version
from flict.flictlib.return_codes import FlictError
from flict.flictlib.return_codes import ReturnCodes
from flict.flictlib import flict_config
from flict.impl import FlictImpl

import sys

DESCRIPTION = """
NAME
  flict (FOSS License Compatibility Tool)

DESCRIPTION
  flict is a Free and Open Source Software tool to verify compatibility between licenses

"""

EPILOG = f"""
Return codes:
{ReturnCodes.get_help()}
CONFIGURATION
  All config files can be found in
  {flict_config.VAR_DIR}

AUTHOR
  Henrik Sandklef

PROJECT SITE
  https://github.com/vinland-technology/flict

REPORTING BUGS
  File a ticket at https://github.com/vinland-technology/flict/issues

COPYRIGHT
  Copyright (c) 2021 Henrik Sandklef<hesa@sandklef.com>.
  License GPL-3.0-or-later

ATTRIBUTION
  flict is using the license compatibility matrix from osadl.org.


"""

OUTPUT_FORMAT_JSON = "JSON"
OUTPUT_FORMAT_TEXT = "text"
OUTPUT_FORMAT_MARKDOWN = "markdown"
OUTPUT_FORMAT_DOT = "dot"

DATE_FMT = '%Y-%m-%d'


def parse():

    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        epilog=EPILOG,
        formatter_class=RawTextHelpFormatter,
    )

    commmon_defaults_group = parser.add_argument_group(title='Options to change default settings')
    deveveloper_group = parser.add_argument_group(title='Developer options')

    # DEFAULTS
    commmon_defaults_group.add_argument('--license-matrix-file', '-lmf',
                                        type=str,
                                        dest='license_matrix_file',
                                        help='File with license compatibility matrix, defaults to osadl-matrix database',
                                        default=flict_config.DEFAULT_MATRIX_FILE)

    commmon_defaults_group.add_argument('--licenses-denied-file', '-ldf', type=str, dest='licenses_denied_file', help='', default=None)

    commmon_defaults_group.add_argument('--licenses-preference-file', '-lpf', type=str, dest='licenses_preference_file', help='', default=None)

    commmon_defaults_group.add_argument('--license-info-file', '-lif', type=str, dest='licenses_info_file', help='Short for applying -lmf <file> -ldf <file> -lpf <file>', default=None)

    # COMMON
    parser.add_argument('-ip', '--ignore-problems',
                        action='store_true',
                        dest='ignore_problems',
                        help='Do not exit with an error code if undefined or unknown licenses are identified.',
                        default=False)

    # COMMON
    parser.add_argument('-el', '--extended-licenses',
                        action='store_true',
                        dest='extended_licenses',
                        help='Check all supported licenes when trying to find an outbound license',
                        default=False)

    # COMMON
    parser.add_argument('-nr', '--no-relicense',
                        action='store_true',
                        dest='no_relicense',
                        help='do not use license relicensing, same as -rf ""',
                        default=False)

    # COMMON
    parser.add_argument('-o', '--output',
                        type=argparse.FileType('w'),
                        dest='output',
                        help='output, defaults to stdout',
                        default=sys.stdout)
    # COMMON
    parser.add_argument('-of', '--output-format',
                        type=str,
                        dest='output_format',
                        help=f'output format. Avilable formats: {OUTPUT_FORMAT_JSON}, {OUTPUT_FORMAT_TEXT}, {OUTPUT_FORMAT_MARKDOWN}, {OUTPUT_FORMAT_DOT}. Defaults to {flict_config.DEFAULT_OUTPUT_FORMAT}',
                        default=flict_config.DEFAULT_OUTPUT_FORMAT)

    parser.add_argument('-v', '--verbose',
                        action='store_true',
                        help='output verbose information to stderr',
                        default=False)

    parser.add_argument('-cc', '--check-compatibility',
                        type=str, nargs='+',
                        dest='licenses',
                        help='licenses to check for compatibility')

    parser.add_argument('-ol', '--outbound-license',
                        type=str,
                        dest='outbound_licenses',
                        help='conclude outbound license suggestions from specified license expression. Example: -ol "GPLv2 and MIT BSD-3"')

    # KEEP
    parser.add_argument('-V', '--version',
                        action='version',
                        version=flict_version,
                        default=False)

    deveveloper_group.add_argument('-dl', '--debug-license',
                                   action='store_true',
                                   dest='debug_license',
                                   help='output verbose debug information of the intermediate steps when transforming a license expression',
                                   default=False)

    subparsers = parser.add_subparsers(help='Sub commands')

    # verify
    parser_v = subparsers.add_parser(
        'verify', help='Verify license compatibility between for a package or an outbound license expression against inbound license expression.')
    parser_v.set_defaults(which="verify", func=verify)
    parser_v.add_argument('--outbound-license', '-ol', type=str, nargs="+", dest='out_license', help='Outbound license expressions', default=None)
    parser_v.add_argument('--inbound-license', '-il', type=str, nargs='+', dest='in_license_expr', help='Inbound license expression', default=[])
    parser_v.add_argument('--sbom', '-s', type=str, dest='verify_sbom', help='SBoM file to verify')
    parser_v.add_argument('--sbom-dirs', '-sd', type=str, nargs='+', dest='sbom_dirs', help='Directories where SBoM files are searched for.', default='.')
    parser_v.add_argument('--flict', '-f', type=str, dest='verify_flict', help='Flict project file to verify')

    parser_v.add_argument('--manifest-file', '-mf', type=str,
                          help='verify license compatibility for project in manifest file')

    # simplify
    parser_si = subparsers.add_parser(
        'simplify', help='Expand and simplify a license expression')
    parser_si.set_defaults(which="simplify", func=simplify)
    parser_si.add_argument('license_expression', type=str, nargs='+',
                           help='License expression to simplify')

    # list
    parser_li = subparsers.add_parser(
        'list', help='List supported licenses')
    parser_li.set_defaults(which="list", func=list_licenses)
    parser_li.add_argument('-t', '--translations',
                           dest='list_translation',
                           action='store_true',
                           help='List translation information')

    merge_parser = subparsers.add_parser('merge', help="Merge additional licenses with OSADL's matrix. Will output to store in a file, for use with --license-matrix-file")
    merge_parser.set_defaults(which="merge", func=_merge_licenses)
    merge_parser.add_argument('--license-file', '-lf', type=str, dest='license_file', help='License file (JSON) to merge', default=None)
    merge_parser.add_argument('--default-no', '-dn', action='store_true', dest='default_no', help='If no compatibility can be found in the additional matrix, "No" compatibility is assumed. Works only for JSON input matrix', default=False)

    # display-compatibility
    parser_d = subparsers.add_parser(
        'display-compatibility', help='Display license compatibility graphically')
    parser_d.set_defaults(which="display-compatibility",
                          func=display_compatibility)
    parser_d.add_argument('license_expression', type=str, nargs='+',
                          help='Licenses, separated by space, to display compatibility for')

    # outbound-candidates
    parser_s = subparsers.add_parser(
        'outbound-candidate', help='Suggest outbound license candidates')
    parser_s.set_defaults(which="outbound-candidate",
                          func=suggest_outbound_candidate)
    parser_s.add_argument('license_expression', type=str, nargs='+',
                          help='License expression to suggest candidate outbound license for')

    # policy-report
    parser_p = subparsers.add_parser(
        'policy-report', help='create report with license policy applied')
    parser_p.set_defaults(which="policy-report", func=policy_report)
    parser_p.add_argument('--license-policy-file', '-lpf',
                          type=str,
                          dest='policy_file',
                          help='file with license policy',
                          required=True)
    parser_p.add_argument('--compliance-report-file', '-crf',
                          type=str,
                          dest='report_file',
                          help='file with report as produced using "verify"',
                          required=True)

    args = parser.parse_args()

    if args.no_relicense:
        args.relicense_file = ""

    return args


def flict_print(args, msg):
    print(msg, file=args.output)


def flict_exit(ret_code, msg=None):
    if msg is not None:
        logger.main_logger.error(msg)
    if isinstance(ret_code, ReturnCodes):
        ret_code = ret_code.value[0]
    exit(ret_code)


def simplify(args):
    ret = FlictImpl(args).simplify()
    flict_print(args, ret)


def _merge_licenses(args):
    file_sanity_check(args.license_file)
    ret = FlictImpl(args).merge_license_db()
    flict_print(args, ret)


def list_licenses(args):
    ret = FlictImpl(args).list_licenses()
    flict_print(args, ret)


def verify(args):
    data, code = FlictImpl(args).verify()
    flict_print(args, data)
    return code


def display_compatibility(args):
    on_error = "Expected at least two licenses for checking compatibility"
    on_error_code = ReturnCodes.RET_MISSING_ARGS
    if len(args.license_expression) == 1:
        flict_exit(on_error_code, on_error)
    ret = FlictImpl(args).display_compatibility()
    flict_print(args, ret)


def suggest_outbound_candidate(args):
    ret = FlictImpl(args).suggest_outbound_candidate()
    flict_print(args, ret)


def policy_report(args):
    file_sanity_check(args.report_file)
    file_sanity_check(args.policy_file)
    ret = FlictImpl(args).policy_report()
    flict_print(args, ret)


def file_sanity_check(fname):
    on_error = "Provided file {fname} was not found, or cannot be read."
    on_error_code = ReturnCodes.RET_FILE_NOT_FOUND
    try:
        open(fname).close()
    except (FileNotFoundError, PermissionError):
        flict_exit(on_error_code, on_error.format(fname=fname))


def main():
    args = parse()
    return_code = ReturnCodes.RET_SUCCESS

    logger.setup(args.debug_license, args.verbose)

    if 'which' in args:
        try:
            fun_ret = args.func(args)
            # Returning None counts as success.  Some functions might
            # return 1 to signal incompatibility, the below is to
            # catch that
            if fun_ret:
                return_code = fun_ret

        except FlictError as e:
            if args.verbose:
                print(traceback.format_exc(), file=sys.stderr)
            flict_exit(e.error_code(), e.error_message())

    else:
        flict_exit(ReturnCodes.RET_MISSING_ARGS, "Missing command.")

    flict_exit(return_code)


if __name__ == '__main__':
    main()
